@@ -5,6 +5,7 @@
 {-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE MagicHash  #-}
 {-# LANGUAGE ScopedTypeVariables #-}
+{-# OPTIONS_GHC -fno-warn-warnings-deprecations #-}
 ---------------------------------------------------------
 --
 -- Module        : Network.Wai.Handler.Warp
@@ -31,6 +32,8 @@ module Network.Wai.Handler.Warp
     , settingsPort
     , settingsHost
     , settingsOnException
+    , settingsOnOpen
+    , settingsOnClose
     , settingsTimeout
     , settingsIntercept
     , settingsManager
@@ -48,7 +51,6 @@ module Network.Wai.Handler.Warp
     , parseRequest
     , sendResponse
     , registerKillThread
-    , bindPort
     , pause
     , resume
     , T.cancel
@@ -56,10 +58,11 @@ module Network.Wai.Handler.Warp
     , T.initialize
 #if TEST
     , takeHeaders
+    , readInt
 #endif
     ) where
 
-import Prelude hiding (catch, lines)
+import Prelude hiding (lines)
 import Network.Wai
 
 import Data.ByteString (ByteString)
@@ -68,29 +71,33 @@ import qualified Data.ByteString.Unsafe as SU
 import qualified Data.ByteString.Char8 as B
 import qualified Data.ByteString.Lazy as L
 import Network (sClose, Socket)
-import Network.Socket
-    ( accept, Family (..)
-    , SocketType (Stream), listen, bindSocket, setSocketOption, maxListenQueue
-    , SockAddr, SocketOption (ReuseAddr)
-    , AddrInfo(..), AddrInfoFlag(..), defaultHints, getAddrInfo
-    )
-import qualified Network.Socket
+import Network.Socket (accept, SockAddr)
 import qualified Network.Socket.ByteString as Sock
 import Control.Exception
-    ( bracket, finally, Exception, SomeException, catch
+    ( mask, handle, onException, bracket
+    , Exception, SomeException
     , fromException, AsyncException (ThreadKilled)
-    , bracketOnError
-    , IOException
     , try
+#if __GLASGOW_HASKELL__ >= 702
+    , allowInterrupt
+#else
+    , unblock
+#endif
+#if WINDOWS
+    , finally
+#endif
     )
 import Control.Concurrent (forkIO)
 import Data.Maybe (fromMaybe, isJust)
+import Data.Char (toLower, isHexDigit)
+import Data.Word (Word)
 
 import Data.Typeable (Typeable)
 
-import Control.Monad.Trans.Resource (ResourceT, runResourceT)
+import Data.Conduit (ResourceT, runResourceT)
 import qualified Data.Conduit as C
 import qualified Data.Conduit.List as CL
+import qualified Data.Conduit.Binary as CB
 import Data.Conduit.Blaze (builderToByteString)
 import Control.Exception.Lifted (throwIO)
 import Blaze.ByteString.Builder.HTTP
@@ -103,18 +110,19 @@ import Network.Sendfile
 
 import qualified System.PosixCompat.Files as P
 
-import Control.Monad.IO.Class (liftIO)
+import Control.Monad.IO.Class (MonadIO, liftIO)
+import Control.Monad.Trans.Class (lift)
 import qualified Timeout as T
 import Timeout (Manager, registerKillThread, pause, resume)
 import Data.Word (Word8)
 import Data.List (foldl')
-import Control.Monad (forever, when)
+import Control.Monad (forever, when, void)
 import qualified Network.HTTP.Types as H
 import qualified Data.CaseInsensitive as CI
 import System.IO (hPrint, stderr)
+import ReadInt (readInt64)
 import qualified Data.IORef as I
-import Data.String (IsString (..))
-import qualified Data.ByteString.Lex.Integral as LI
+import Data.Conduit.Network (bindPort, HostPreference (HostIPv4))
 
 #if WINDOWS
 import Control.Concurrent (threadDelay)
@@ -128,6 +136,11 @@ import qualified Paths_warp
 warpVersion :: String
 warpVersion = showVersion Paths_warp.version
 
+#if __GLASGOW_HASKELL__ < 702
+allowInterrupt :: IO ()
+allowInterrupt = unblock $ return ()
+#endif
+
 -- |
 --
 -- In order to provide slowloris protection, Warp provides timeout handlers. We
@@ -162,47 +175,6 @@ socketConnection s = Connection
     , connRecv = Sock.recv s bytesPerRead
     }
 
-
-bindPort :: Int -> HostPreference -> IO Socket
-bindPort p s = do
-    let hints = defaultHints { addrFlags = [AI_PASSIVE
-                                         , AI_NUMERICSERV
-                                         , AI_NUMERICHOST]
-                             , addrSocketType = Stream }
-        host =
-            case s of
-                Host s' -> Just s'
-                _ -> Nothing
-        port = Just . show $ p
-    addrs <- getAddrInfo (Just hints) host port
-    -- Choose an IPv6 socket if exists.  This ensures the socket can
-    -- handle both IPv4 and IPv6 if v6only is false.
-    let addrs4 = filter (\x -> addrFamily x /= AF_INET6) addrs
-        addrs6 = filter (\x -> addrFamily x == AF_INET6) addrs
-        addrs' =
-            case s of
-                HostIPv4 -> addrs4 ++ addrs6
-                HostIPv6 -> addrs6 ++ addrs4
-                _ -> addrs
-
-        tryAddrs (addr1:rest@(_:_)) = 
-                                      catch
-                                      (theBody addr1) 
-                                      (\(_ :: IOException) -> tryAddrs rest)
-        tryAddrs (addr1:[])         = theBody addr1
-        tryAddrs _                  = error "bindPort: addrs is empty"
-        theBody addr = 
-          bracketOnError
-          (Network.Socket.socket (addrFamily addr) (addrSocketType addr) (addrProtocol addr))
-          sClose
-          (\sock -> do
-              setSocketOption sock ReuseAddr 1
-              bindSocket sock (addrAddress addr)
-              listen sock maxListenQueue
-              return sock
-          )
-    tryAddrs addrs'
-
 -- | Run an 'Application' on the given port. This calls 'runSettings' with
 -- 'defaultSettings'.
 run :: Port -> Application -> IO ()
@@ -249,34 +221,92 @@ runSettingsConnection :: Settings -> IO (Connection, SockAddr) -> Application ->
 runSettingsConnection set getConn app = do
     let onE = settingsOnException set
         port = settingsPort set
+        onOpen = settingsOnOpen set
+        onClose = settingsOnClose set
     tm <- maybe (T.initialize $ settingsTimeout set * 1000000) return
         $ settingsManager set
-    forever $ do
+    mask $ \restore -> forever $ do
+        allowInterrupt
         (conn, addr) <- getConn
-        _ <- forkIO $ do
+        void $ forkIO $ do
             th <- T.registerKillThread tm
-            serveConnection set th onE port app conn addr
-            T.cancel th
-        return ()
+            handle onE $ (do onOpen
+                             restore $ serveConnection set th port app conn addr
+                             connClose conn >> T.cancel th >> onClose
+                         ) `onException` (T.cancel th >> connClose conn >> onClose)
+
+-- | Contains a @Source@ and a byte count that is still to be read in.
+newtype IsolatedBSSource = IsolatedBSSource (I.IORef (Int, C.Source (ResourceT IO) ByteString))
+
+-- | Given an @IsolatedBSSource@ provide a @Source@ that only allows up to the
+-- specified number of bytes to be passed downstream. All leftovers should be
+-- retained within the @Source@. If there are not enough bytes available,
+-- throws a @ConnectionClosedByPeer@ exception.
+ibsIsolate :: IsolatedBSSource -> C.Source (ResourceT IO) ByteString
+ibsIsolate ibs@(IsolatedBSSource ref) =
+    C.PipeM pull (return ())
+  where
+    pull = do
+        (count, src) <- liftIO $ I.readIORef ref
+        if count == 0
+            -- No more bytes wanted downstream, so we're done.
+            then return $ C.Done Nothing ()
+            else do
+                -- Get the next chunk (if available) and the updated source
+                (src', mbs) <- src C.$$+ CL.head
+
+                -- If no chunk available, then there aren't enough bytes in the
+                -- stream. Throw a ConnectionClosedByPeer
+                bs <- maybe (liftIO $ throwIO ConnectionClosedByPeer) return mbs
+
+                let -- How many of the bytes in this chunk to send downstream
+                    toSend = min count (S.length bs)
+                    -- How many bytes will still remain to be sent downstream
+                    count' = count - toSend
+                case () of
+                    ()
+                        -- The expected count is greater than the size of the
+                        -- chunk we just read. Send the entire chunk
+                        -- downstream, and then loop on this function for the
+                        -- next chunk.
+                        | count' > 0 -> do
+                            liftIO $ I.writeIORef ref (count', src')
+                            return $ C.HaveOutput (ibsIsolate ibs) (return ()) bs
+
+                        -- The expected count is the total size of the chunk we
+                        -- just read. Send this chunk downstream, and then
+                        -- terminate the stream.
+                        | count == S.length bs -> do
+                            liftIO $ I.writeIORef ref (count', src')
+                            return $ C.HaveOutput (C.Done Nothing ()) (return ()) bs
+
+                        -- Some of the bytes in this chunk should not be sent
+                        -- downstream. Split up the chunk into the sent and
+                        -- not-sent parts, add the not-sent parts onto the new
+                        -- source, and send the rest of the chunk downstream.
+                        | otherwise -> do
+                            let (x, y) = S.splitAt toSend bs
+                            liftIO $ I.writeIORef ref (count', C.HaveOutput src' (return ()) y)
+                            return $ C.HaveOutput (C.Done Nothing ()) (return ()) x
+
+-- | Extract the underlying @Source@ from an @IsolatedBSSource@, which will not
+-- perform any more isolation.
+ibsDone :: IsolatedBSSource -> IO (C.Source (ResourceT IO) ByteString)
+ibsDone (IsolatedBSSource ref) = fmap snd $ I.readIORef ref
 
 serveConnection :: Settings
                 -> T.Handle
-                -> (SomeException -> IO ())
-                -> Port -> Application -> Connection -> SockAddr -> IO ()
-serveConnection settings th onException port app conn remoteHost' =
-    catch
-        (finally
-          (runResourceT serveConnection')
-          (connClose conn))
-        onException
+                -> Port -> Application -> Connection -> SockAddr-> IO ()
+serveConnection settings th port app conn remoteHost' =
+    runResourceT serveConnection'
   where
     serveConnection' :: ResourceT IO ()
     serveConnection' = do
-        fromClient <- C.bufferSource $ connSource conn th
+        let fromClient = connSource conn th
         serveConnection'' fromClient
 
     serveConnection'' fromClient = do
-        env <- parseRequest conn port remoteHost' fromClient
+        (env, getSource) <- parseRequest conn port remoteHost' fromClient
         case settingsIntercept settings env of
             Nothing -> do
                 -- Let the application run for as long as it wants
@@ -285,20 +315,22 @@ serveConnection settings th onException port app conn remoteHost' =
 
                 -- flush the rest of the request body
                 requestBody env C.$$ CL.sinkNull
+                fromClient' <- liftIO getSource
 
                 liftIO $ T.resume th
                 keepAlive <- sendResponse th env conn res
-                when keepAlive $ serveConnection'' fromClient
+                when keepAlive $ serveConnection'' fromClient'
             Just intercept -> do
                 liftIO $ T.pause th
-                intercept fromClient conn
+                fromClient' <- liftIO getSource
+                intercept fromClient' conn
 
 parseRequest :: Connection -> Port -> SockAddr
-             -> C.BufferedSource IO S.ByteString
-             -> ResourceT IO Request
-parseRequest conn port remoteHost' src = do
-    headers' <- src C.$$ takeHeaders
-    parseRequest' conn port headers' remoteHost' src
+             -> C.Source (ResourceT IO) S.ByteString
+             -> ResourceT IO (Request, IO (C.Source (ResourceT IO) ByteString))
+parseRequest conn port remoteHost' src1 = do
+    (src2, headers') <- src1 C.$$+ takeHeaders
+    parseRequest' conn port headers' remoteHost' src2
 
 -- FIXME come up with good values here
 bytesPerRead, maxTotalHeaderLength :: Int
@@ -310,6 +342,7 @@ data InvalidRequest =
     | BadFirstLine String
     | NonHttp
     | IncompleteHeaders
+    | ConnectionClosedByPeer
     | OverLargeHeader
     deriving (Show, Typeable, Eq)
 instance Exception InvalidRequest
@@ -333,8 +366,8 @@ parseRequest' :: Connection
               -> Port
               -> [ByteString]
               -> SockAddr
-              -> C.BufferedSource IO S.ByteString
-              -> ResourceT IO Request
+              -> C.Source (ResourceT IO) S.ByteString -- FIXME was buffered
+              -> ResourceT IO (Request, IO (C.Source (ResourceT IO) ByteString))
 parseRequest' _ _ [] _ _ = throwIO $ NotEnoughLines []
 parseRequest' conn port (firstLine:otherLines) remoteHost' src = do
     (method, rpath', gets, httpversion) <- parseFirst firstLine
@@ -349,60 +382,20 @@ parseRequest' conn port (firstLine:otherLines) remoteHost' src = do
     let len0 =
             case lookup "content-length" heads of
                 Nothing -> 0
-                Just bs -> LI.readDecimal_ bs
+                Just bs -> readInt bs
     let serverName' = takeUntil 58 host -- ':'
-    rbody <-
-        if len0 == 0
-            then return mempty
-            else do
-                -- We can't use the standard isolate, as its counter is not
-                -- kept in a mutable variable.
-                lenRef <- liftIO $ I.newIORef len0
-                let isolate = C.Conduit push close
-                    push bs = do
-                        len <- liftIO $ I.readIORef lenRef
-                        let (a, b) = S.splitAt len bs
-                            len' = len - S.length a
-                        liftIO $ I.writeIORef lenRef len'
-                        return $ if len' == 0
-                            then C.Finished (if S.null b then Nothing else Just b) (if S.null a then [] else [a])
-                            else C.Producing isolate [a]
-                    close = return []
+    let chunked = maybe False ((== "chunked") . B.map toLower)
+                  $ lookup "transfer-encoding" heads
+    (rbody, getSource) <- liftIO $
+        if chunked
+          then do
+            ref <- I.newIORef (src, NeedLen)
+            return (chunkedSource ref, fmap fst $ I.readIORef ref)
+          else do
+            ibs <- fmap IsolatedBSSource $ I.newIORef (len0, src)
+            return (ibsIsolate ibs, ibsDone ibs)
 
-                    -- Make sure that we don't connect to the source after the
-                    -- isolate conduit closes.
-                    --
-                    -- Here's the issue: we fuse our buffered request body with
-                    -- an isolate conduit which ensures no more than X bytes
-                    -- are read. Suppose we read all X bytes, and then we call
-                    -- requestBody again. What happens?
-                    --
-                    -- Previously, we would try to read one more chunk from the
-                    -- buffered source. This is inherent to conduit: we
-                    -- wouldn't know that the isolate Conduit isn't accepting
-                    -- more data until after we've pushed some data to it. This
-                    -- results in hanging, since there's no data available on
-                    -- the wire.
-                    --
-                    -- Instead, we add a wrapper that checks if the request
-                    -- body has already been depleted before making that first
-                    -- pull.
-                    --
-                    -- Possible optimization: do away with the Conduit
-                    -- entirely. However, this may be less efficient overall,
-                    -- as we'd now have to check the BufferedSource status on
-                    -- each call. Worth looking into.
-
-                    wrap src' = C.Source
-                        { C.sourcePull = do
-                            len <- liftIO $ I.readIORef lenRef
-                            if len <= 0
-                                then return C.Closed
-                                else C.sourcePull src'
-                        , C.sourceClose = return ()
-                        }
-                return $ wrap $ src C.$= isolate
-    return Request
+    return (Request
             { requestMethod = method
             , httpVersion = httpversion
             , pathInfo = H.decodePathSegments rpath
@@ -416,8 +409,73 @@ parseRequest' conn port (firstLine:otherLines) remoteHost' src = do
             , remoteHost = remoteHost'
             , requestBody = rbody
             , vault = mempty
-            }
+            }, getSource)
 
+data ChunkState = NeedLen
+                | NeedLenNewline
+                | HaveLen Word
+
+chunkedSource :: MonadIO m
+              => I.IORef (C.Source m ByteString, ChunkState)
+              -> C.Source m ByteString
+chunkedSource ipair = do
+    (src, mlen) <- liftIO $ I.readIORef ipair
+    go src mlen
+  where
+    go' src front = do
+        (src', (len, bs)) <- lift $ src C.$$+ front getLen
+        let src''
+                | S.null bs = src'
+                | otherwise = C.yield bs >> src'
+        go src'' $ HaveLen len
+
+    go src NeedLen = go' src id
+    go src NeedLenNewline = go' src (CB.take 2 >>)
+    go src (HaveLen 0) = liftIO $ I.writeIORef ipair (src, HaveLen 0)
+    go src (HaveLen len) = do
+        (src', mbs) <- lift $ src C.$$+ CL.head
+        case mbs of
+            Nothing -> liftIO $ I.writeIORef ipair (src', HaveLen 0)
+            Just bs ->
+                case S.length bs `compare` fromIntegral len of
+                    EQ -> yield' src' NeedLenNewline bs
+                    LT -> do
+                        let mlen = HaveLen $ len - fromIntegral (S.length bs)
+                        yield' src' mlen bs
+                    GT -> do
+                        let (x, y) = S.splitAt (fromIntegral len) bs
+                        let src'' = C.yield y >> src'
+                        yield' src'' NeedLenNewline x
+
+    yield' src mlen bs = do
+        liftIO $ I.writeIORef ipair (src, mlen)
+        C.yield bs
+        go src mlen
+
+    getLen :: Monad m => C.Sink ByteString m (Word, ByteString)
+    getLen = do
+        mbs <- CL.head
+        case mbs of
+            Nothing -> return (0, S.empty)
+            Just bs -> do
+                (x, y) <-
+                    case S.breakByte 10 bs of
+                        (x, y)
+                            | S.null y -> do
+                                mbs2 <- CL.head
+                                case mbs2 of
+                                    Nothing -> return (x, y)
+                                    Just bs2 -> return $ S.breakByte 10 $ bs `S.append` bs2
+                            | otherwise -> return (x, y)
+                let w =
+                        S.foldl' (\i c -> i * 16 + fromIntegral (hexToWord c)) 0
+                        $ B.takeWhile isHexDigit x
+                return (w, S.drop 1 y)
+
+    hexToWord w
+        | w < 58 = w - 48
+        | w < 71 = w - 55
+        | otherwise = w - 87
 
 takeUntil :: Word8 -> ByteString -> ByteString
 takeUntil c bs =
@@ -512,7 +570,7 @@ sendResponse th req conn r = sendResponse' r
     sendResponse' :: Response -> ResourceT IO Bool
     sendResponse' (ResponseFile s hs fp mpart) = do
         eres <-
-            case (LI.readDecimal_ `fmap` lookup "content-length" hs, mpart) of
+            case (readInt `fmap` lookup "content-length" hs, mpart) of
                 (Just cl, _) -> return $ Right (hs, cl)
                 (Nothing, Nothing) -> liftIO $ try $ do
                     cl <- P.fileSize `fmap` P.getFileStatus fp
@@ -570,21 +628,23 @@ sendResponse th req conn r = sendResponse' r
             T.tickle th
             return isPersist
       where
-        body = fmap (\x -> case x of
+        body = fmap2 (\x -> case x of
                         C.Flush -> flush
                         C.Chunk builder -> builder) bodyFlush
         headers' = headers version s hs
         -- FIXME perhaps alloca a buffer per thread and reuse that in all
         -- functions below. Should lessen greatly the GC burden (I hope)
         needsChunked' = needsChunked hs
-        chunk :: C.Conduit Builder IO Builder
-        chunk = C.Conduit
-            { C.conduitPush = push
-            , C.conduitClose = close
-            }
-        conduit = C.Conduit push close
-        push x = return $ C.Producing conduit [chunkedTransferEncoding x]
-        close = return [chunkedTransferTerminator]
+        chunk :: C.Conduit Builder (ResourceT IO) Builder
+        chunk = C.NeedInput push close
+        push x = C.HaveOutput chunk (return ()) (chunkedTransferEncoding x)
+        close = C.HaveOutput (C.Done Nothing ()) (return ()) chunkedTransferTerminator
+
+fmap2 :: Functor m => (o1 -> o2) -> C.Pipe i o1 m r -> C.Pipe i o2 m r
+fmap2 f (C.HaveOutput p c o) = C.HaveOutput (fmap2 f p) c (f o)
+fmap2 f (C.NeedInput p c) = C.NeedInput (fmap2 f . p) (fmap2 f c)
+fmap2 f (C.PipeM mp c) = C.PipeM (fmap (fmap2 f) mp) c
+fmap2 _ (C.Done i x) = C.Done i x
 
 parseHeaderNoAttr :: ByteString -> H.Header
 parseHeaderNoAttr s =
@@ -596,33 +656,31 @@ parseHeaderNoAttr s =
                    else rest
      in (CI.mk k, rest')
 
-connSource :: Connection -> T.Handle -> C.Source IO ByteString
+connSource :: Connection -> T.Handle -> C.Source (ResourceT IO) ByteString
 connSource Connection { connRecv = recv } th =
     src
   where
-    src = C.Source
-        { C.sourcePull = do
-            bs <- liftIO recv
-            if S.null bs
-                then return C.Closed
-                else do
-                    when (S.length bs >= 2048) $ liftIO $ T.tickle th
-                    return (C.Open src bs)
-        , C.sourceClose = return ()
-        }
+    src = C.PipeM (do
+        bs <- liftIO recv
+        if S.null bs
+            then return $ C.Done Nothing ()
+            else do
+                when (S.length bs >= 2048) $ liftIO $ T.tickle th
+                return (C.HaveOutput src (return ()) bs))
+        (return ())
 
 -- | Use 'connSendAll' to send this data while respecting timeout rules.
-connSink :: Connection -> T.Handle -> C.Sink B.ByteString IO ()
+connSink :: Connection -> T.Handle -> C.Sink B.ByteString (ResourceT IO) ()
 connSink Connection { connSendAll = send } th =
-    C.SinkData push close
+    sink
   where
+    sink = C.NeedInput push close
     close = liftIO (T.resume th)
-    push x = do
-        liftIO $ do
-            T.resume th
-            send x
-            T.pause th
-        return (C.Processing push close)
+    push x = C.PipeM (liftIO $ do
+        T.resume th
+        send x
+        T.pause th
+        return sink) (liftIO $ T.resume th)
     -- We pause timeouts before passing control back to user code. This ensures
     -- that a timeout will only ever be executed when Warp is in control. We
     -- also make sure to resume the timeout after the completion of user code
@@ -641,43 +699,13 @@ data Settings = Settings
     { settingsPort :: Int -- ^ Port to listen on. Default value: 3000
     , settingsHost :: HostPreference -- ^ Default value: HostIPv4
     , settingsOnException :: SomeException -> IO () -- ^ What to do with exceptions thrown by either the application or server. Default: ignore server-generated exceptions (see 'InvalidRequest') and print application-generated applications to stderr.
+    , settingsOnOpen :: IO () -- ^ What to do when a connection is open. Default: do nothing.
+    , settingsOnClose :: IO ()  -- ^ What to do when a connection is close. Default: do nothing.
     , settingsTimeout :: Int -- ^ Timeout value in seconds. Default value: 30
-    , settingsIntercept :: Request -> Maybe (C.BufferedSource IO S.ByteString -> Connection -> ResourceT IO ())
+    , settingsIntercept :: Request -> Maybe (C.Source (ResourceT IO) S.ByteString -> Connection -> ResourceT IO ())
     , settingsManager :: Maybe Manager -- ^ Use an existing timeout manager instead of spawning a new one. If used, 'settingsTimeout' is ignored. Default is 'Nothing'
     }
 
--- | Which host to bind.
---
--- Note: The @IsString@ instance recognizes the following special values:
---
--- * @*@ means @HostAny@
---
--- * @*4@ means @HostIPv4@
---
--- * @*6@ means @HostIPv6@
-data HostPreference =
-    HostAny
-  | HostIPv4
-  | HostIPv6
-  | Host String
-    deriving (Show, Eq, Ord)
-
-instance IsString HostPreference where
-    -- The funny code coming up is to get around some irritating warnings from
-    -- GHC. I should be able to just write:
-    {-
-    fromString "*" = HostAny
-    fromString "*4" = HostIPv4
-    fromString "*6" = HostIPv6
-    -}
-    fromString s'@('*':s) =
-        case s of
-            [] -> HostAny
-            ['4'] -> HostIPv4
-            ['6'] -> HostIPv6
-            _ -> Host s'
-    fromString s = Host s
-
 -- | The default settings for the Warp server. See the individual settings for
 -- the default value.
 defaultSettings :: Settings
@@ -690,6 +718,8 @@ defaultSettings = Settings
             Nothing ->
                 when (go' $ fromException e) $
                     hPrint stderr e
+    , settingsOnOpen = return ()
+    , settingsOnClose = return ()
     , settingsTimeout = 30
     , settingsIntercept = const Nothing
     , settingsManager = Nothing
@@ -708,52 +738,56 @@ data THStatus = THStatus
     BSEndoList -- previously parsed lines
     BSEndo -- bytestrings to be prepended
 
-takeHeaders :: C.Sink ByteString IO [ByteString]
+takeHeaders :: C.Sink ByteString (ResourceT IO) [ByteString]
 takeHeaders =
-    C.sinkState (THStatus 0 id id) takeHeadersPush close
+    C.NeedInput (push (THStatus 0 id id)) close
   where
-    close _ = throwIO IncompleteHeaders
-{-# INLINE takeHeaders #-}
+    close = throwIO IncompleteHeaders
 
-takeHeadersPush :: THStatus
-                -> ByteString
-                -> ResourceT IO (C.SinkStateResult THStatus ByteString [ByteString])
-takeHeadersPush (THStatus len _ _ ) _
-    | len > maxTotalHeaderLength = throwIO OverLargeHeader
-takeHeadersPush (THStatus len lines prepend) bs =
-    case mnl of
-        -- no newline.  prepend entire bs to next line
-        Nothing -> do
-            let len' = len + bsLen
-            return $ C.StateProcessing $ THStatus len' lines (prepend . S.append bs)
-        Just nl -> do
-            let end = nl
-                start = nl + 1
-                line = prepend (if end > 0
-                                    then SU.unsafeTake (checkCR bs end) bs
-                                    else S.empty)
-            if S.null line
-                -- no more headers
-                then do
-                    let lines' = lines []
-                    if start < bsLen
-                        then do
-                            let rest = SU.unsafeDrop start bs
-                            return $ C.StateDone (Just rest) lines'
-                        else return $ C.StateDone Nothing lines'
-                -- more headers
-                else do
-                    let len' = len + start
-                        lines' = lines . (:) line
-                    if start < bsLen
-                        then do
-                            let more = SU.unsafeDrop start bs
-                            takeHeadersPush (THStatus len' lines' id) more
-                        else return $ C.StateProcessing $ THStatus len' lines' id
-  where
-    bsLen = S.length bs
-    mnl = S.elemIndex 10 bs
-{-# INLINE takeHeadersPush #-}
+    push (THStatus len lines prepend) bs
+        -- Too many bytes
+        | len > maxTotalHeaderLength = throwIO OverLargeHeader
+        | otherwise =
+            case mnl of
+                -- No newline find in this chunk.  Add it to the prepend,
+                -- update the length, and continue processing.
+                Nothing ->
+                    let len' = len + bsLen
+                        prepend' = prepend . S.append bs
+                        status = THStatus len' lines prepend'
+                     in C.NeedInput (push status) close
+                -- Found a newline at position end.
+                Just end ->
+                    let start = end + 1 -- start of next chunk
+                        line
+                            -- There were some bytes before the newline, get them
+                            | end > 0 = prepend $ SU.unsafeTake (checkCR bs end) bs
+                            -- No bytes before the newline
+                            | otherwise = prepend S.empty
+                     in if S.null line
+                            -- no more headers
+                            then
+                                let lines' = lines []
+                                    -- leftover
+                                    rest = if start < bsLen
+                                               then Just (SU.unsafeDrop start bs)
+                                               else Nothing
+                                 in C.Done rest lines'
+                            -- more headers
+                            else
+                                let len' = len + start
+                                    lines' = lines . (line:)
+                                    status = THStatus len' lines' id
+                                 in if start < bsLen
+                                        -- more bytes in this chunk, push again
+                                        then let bs' = SU.unsafeDrop start bs
+                                              in push status bs'
+                                        -- no more bytes in this chunk, ask for more
+                                        else C.NeedInput (push status) close
+      where
+        bsLen = S.length bs
+        mnl = S.elemIndex 10 bs
+{-# INLINE takeHeaders #-}
 
 checkCR :: ByteString -> Int -> Int
 checkCR bs pos =
@@ -763,6 +797,11 @@ checkCR bs pos =
         else pos
 {-# INLINE checkCR #-}
 
+readInt :: Integral a => ByteString -> a
+readInt bs = fromIntegral $ readInt64 bs
+{-# INLINE readInt #-}
+
+
 -- | Call the inner function with a timeout manager.
 withManager :: Int -- ^ timeout in microseconds
             -> (Manager -> IO a)
